;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: SNEPS; Base: 10 -*-

;; Copyright (C) 1984, 1988, 1989, 1993 Research Foundation of 
;;                                      State University of New York

;; Version: $Id: nodeset.lisp,v 1.4 1993/06/04 06:24:00 snwiz Exp $

;; This file is part of SNePS.

;; SNePS is free software; you may redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; SNePS is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with SNePS; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA, or to
;; Dr. Stuart C. Shapiro, Department of Computer Science, State University of
;; New York at Buffalo, 226 Bell Hall, Buffalo, NY 14260, USA

(in-package :sneps)


; =============================================================================
; Data Type:  <node set> : (<node> <node> ... <node>)  
; =============================================================================
;
; Implementation Note: A <node set> is implemented as an ordered list of <node>s
;                      using isless.n as the ordering function.
; 
; =============================================================================
;
; new.ns 
; ------
;
;       returns       : <node set>
;
;       description   : creates a <newnodeset>
;
;                                        written:  rmo 07/28/83
;                                        translated: robert quinn 02/18/86
;                                        modified: scs 06/06/87
;
;
(defmacro new.ns ()
  "Returns a new, empty node set."
  nil)
;
;
; =============================================================================
;
; isnew.ns 
; --------
;
;       arguments     : ns - <node set>
;
;       returns       : <boolean>
;
;       description   : returns "true" if "ns" is a new <node set>,
;                       "false" otherwise.
;
;                                        written:  rmo 07/28/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defmacro isnew.ns (ns)
  "Returns T if NS is a new, empty node set, NIL otherwise."
   `(null ,ns))
;
;
; =============================================================================
;
; is.ns
; -----
;
;       arguments     : u - <universal>
;
;       returns       : <boolean> 
;
;       description   : returns "true" if "u" is a <node set>, "false" otherwise
;
;       implementation: it does not check if all the elements of "u" are <node>s
;                       it checks just the first one.
;
;                                        written : rmo 07/28/83 
;                                        modified: ejm 08/30/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun is.ns (u)
  "Returns T if U is a node set, NIL otherwise.  Only checks first member."
  (and (listp u)
       (or (isnew.ns u)
	   (is.n (first u)))))
;
;
; =============================================================================
;
; isall.ns
; --------
;
;       arguments     : u - <universal>
;
;       returns       : <boolean>
;
;       description   : It returns "true" if "u" is a <node set>,
;                       "false" otherwise.
;
;       implementation: Unlike "is.ns", "isall.ns" tests whether all the 
;                       elements of "u" are <node>s.
;
;                                        written :  ejm 06/22/84
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun isall.ns (u)
  "Returns T if U is a node set, NIL otherwise. Checks everything."
  (and (listp u) 
       (isall1.ns u)))
;
;
; =============================================================================
;
; isall1.ns
; ---------
;
;       arguments     : s - <sequence>
;
;       returns       : <boolean>
;
;       description   : Auxiliary function to isall.ns. It tests whether
;                       "s" is a <node set> by checking whether all
;                       the elements of "s" are <node>s, and if so,
;                       whether they are ordered.
;
;                                        written :  ejm 06/22/84
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;

(defun isall1.ns (s)
  "Returns T if the non-empty list S is a node set, NIL otherwise."
  (cond ((null s) t)
	((not (is.n (first s))) nil)
	((null (rest s)) t)
	((and (is.n (second s)) (isless.n (first s) (second s)))
	 (isall1.ns (rest s)))
	(t nil)))
;
;
; =============================================================================
;
; makeone.ns
; ----------
;
;       arguments     : n - <node>
;
;       returns       : <node set>
;
;       description   : returns a <node set> composed of the <node>.
;
;       note          : This function should be used whenever one wants to
;                       create a <node set> with just one <node>.
;                       Don't use the alternative forms
;
;                                 (make.ns n)
;                       or
;
;                                 (insert.ns n (new.ns))
;
;                       in order to improve efficiency.
;
;                                        written :  ejm 10/20/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defmacro makeone.ns (n)
  "Returns a node set containing just the node N."
  `(list ,n))
;
;
; =============================================================================
;
; insert.ns 
; ---------
;
;       arguments     : n - <node>
;                       ns - <node set>
;
;       returns       : <node set> 
;
;       description   : returns a <node set> identical to "ns" but with "n" as
;                       a new <node> if "n" was not yet in "ns". 
;                       If "n" was in "ns" it just returns "ns" unchanged.
;
;                                        written:  rmo 07/28/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;                                                  scs 12/23/88
;                                                  ssc 03/15/89
;                                                  ssc  5/10/89
;
;
(defun insert.ns (n ns)
  "Returns a node set like NS, but with N inserted."
  (if (member n ns)
      ns
      (let ((tl (member n ns :test #'isless.n)))
	(if tl
	    (nconc (ldiff ns tl) (list n) tl)
	    (append ns (list n))))))

;
;
; =============================================================================
;
; make.ns
; -------
;
;       arguments     : n ... n - <node>s
;
;       returns       : <node set>
;
;       description   : returns a <node set> composed of the <node>s passed as
;                       arguments.
;
;                                        written :  ejm 09/19/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;                                        modified: scs 12/23/88
;
;
(defmacro make.ns (&rest n)
  "Returns a node set composed of the argument nodes."
  (cond ((null n) (new.ns))
	(t `(insert.ns ,(first n) (make.ns ,@(rest n))))))
;
;
; =============================================================================
;
; union.ns 
; --------
;
;       arguments     : ns1 - <node set>
;                       ns2 - <node set>
;
;       returns       : <node set>
;
;       description   : returns the union of "ns1" and "ns2"
;
;                                        written:  rmo 07/28/63
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun union.ns (ns1 ns2)
  "Returns the union of the two node sets, NS1 and NS2."
  (let* ((result (list nil)) (last result))
    (loop (cond ((null ns1) (rplacd last ns2) (return))
		((null ns2) (rplacd last ns1) (return))
		((isless.n (first ns1) (first ns2))
		 (rplacd last (list (first ns1)))
		 (setq last (rest last))
		 (setq ns1 (rest ns1)))
		((isless.n (first ns2) (first ns1))
		 (rplacd last (list (first ns2)))
		 (setq last (rest last))
		 (setq ns2 (rest ns2)))
		(t (rplacd last (list (first ns1)))
		   (setq last (rest last))
		   (setq ns1 (rest ns1))
		   (setq ns2 (rest ns2)))))
    (rest result)))

;
;
; =============================================================================
;
; remove.ns 
; ---------
;
;       arguments     : n - <node>
;                       ns - <node set>
;
;       returns       : <node set> 
;
;       description   : returns a <node set> identical to "ns" but with "n"
;                       removed if it was there. If "n" was not in "ns"
;                       it just returns "ns" unchanged.
;
;                                        written:  ejm 08/15/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun remove.ns (n ns)
  "Returns a node set like NS, but with the node N absent."
  (cond ((isnew.ns ns) ns)
	((isless.n n (first ns)) ns)
	((iseq.n n (first ns)) (rest ns))
	(t (cons (first ns) (remove.ns n (rest ns))))))
;
;
; =============================================================================
;
; choose.ns 
; ---------
;
;       arguments     : ns - <node set>
;
;       returns       : <node> 
;
;       description   : returns a <node> of "ns"
;
;                                        written:  rmo 07/28/63
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun choose.ns (ns)
  "Returns an arbitrary, but deterministic member of the node set NS."
  (first ns))
;
;
; =============================================================================
;
; others.ns 
; ---------
;
;       arguments     : ns - <node set>
;
;       returns       : <node set>
;
;       description   : returns a <node set> identical to "ns" but without
;                       the element that would be chosen by choose.ns
;
;                                        written:  rmo 07/28/63
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun others.ns (ns)
  "Returns a node set like NS, but with (CHOOSE.NS NS) deleted."
   (rest ns))
;
;
; =============================================================================

(defmacro do.ns ((var nsform &optional resultform) &body forms)
  `(dolist (,var ,nsform ,resultform) ,@forms))

; =============================================================================
;
; intersect.ns   
; ------------
;
;       arguments     : ns1 - <node set>
;                       ns2 - <node set>
;
;       returns       : <node set>
;
;       description   : returns the intersection of "ns1" and "ns2"
;
;
;                                        written:  rmo 08/15/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun intersect.ns (ns1 ns2)
  "Returns the intersection of the node sets NS1 and NS2."
 (let* ((result (list nil)) (last result))
    (loop (cond ((null ns1) (return))
		((null ns2) (return))
		((isless.n (first ns1) (first ns2))
		 (setq ns1 (rest ns1)))
		((isless.n (first ns2) (first ns1))
		 (setq ns2 (rest ns2)))
		(t (rplacd last (list (first ns1)))
		   (setq last (rest last))
		   (setq ns1 (rest ns1))
		   (setq ns2 (rest ns2)))))
    (rest result)))

;
;
; =============================================================================
;
; compl.ns 
; -------
;
;       arguments     : ns1 - <node set>
;                       ns2 - <node set> 
;
;       returns       : <node set>
;
;       description   : returns the set difference ns1 - ns2. 
;
;                                        written:  rmo 07/28/83
;                                        modified: ejm 08/19/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun compl.ns (ns1 ns2)
  "Returns the relative complement of node set NS1 and node set NS2."
 (let* ((result (list nil)) (last result))
    (loop (cond ((null ns1) (return))
		((null ns2) (rplacd last ns1) (return))
		((isless.n (first ns1) (first ns2))
		 (rplacd last (list (first ns1)))
		 (setq last (rest last))
		 (setq ns1 (rest ns1)))
		((isless.n (first ns2) (first ns1))
		 (setq ns2 (rest ns2)))
		(t (setq ns1 (rest ns1))
		   (setq ns2 (rest ns2)))))
    (rest result)))

;
;
; =============================================================================
;
; ismemb.ns    
; ---------
;
;       arguments     : n - <node>
;                       ns - <node set> 
;
;       returns       : <boolean> 
;
;       description   : returns "true" if "n"  is an element of "ns",
;                       "false" otherwise.
;
;                                        written:  rmo 07/28/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun ismemb.ns (n ns)
  "Returns T if node N is in the node set NS."
  (cond ((null ns) nil)
	((isless.n n (first ns)) nil)
	((iseq.n n (first ns)) t)
	(t (member n (rest ns) :test #'iseq.n))))
;
;
; =============================================================================
;
; issubset.ns
; -----------
;
;       arguments     : ns1 - <node set>
;                       ns2 - <node set>
;
;       returns       : <boolean>
;
;       description   : returns "true" if "ns1" is a subset of "ns2", 
;                       "false" otherwise.
;
;                                        written:  rmo 07/28/83
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun issubset.ns (ns1 ns2)
  "Returns T if node set NS1 is a subset of node set NS2; NIL otherwise."
  (loop
    (cond ((null ns1) (return t))
	  ((null ns2) (return nil))
	  ((isless.n (first ns1) (first ns2)) (return nil))
	  ((isless.n (first ns2) (first ns1)) (setq ns2 (rest ns2)))
	  (t (setq ns1 (rest ns1)
		   ns2 (rest ns2))))))
	
;
;
; =============================================================================
;
; iseq.ns 
; -------
;
;       arguments     : ns1 - <node set>
;                       ns2 - <node set>
;
;       returns       : <boolean>
;
;       description   : Compares "ns1" and "ns2" as sets : returns "true" if
;                       "ns1" and "ns2" have exactly the same elements,
;                       "false" otherwise.
;
;                                        written:  rmo 07/28/83 
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defmacro iseq.ns (ns1 ns2)
  "Returns T if node sets NS1 and NS2 are equal; NIL otherwise."
  `(equal ,ns1 ,ns2))
;
;
; =============================================================================
;
; describe.ns
; ------------
;
;       arguments     : ns - <node set>
;
;       returns       : <sequence>
;
;       description   : It returns a <sequence> which is a description of 
;                       the <node set> "ns" to be printed.
;
;                                        written :  ejm 06/05/84
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
(defmacro describe.ns (ns)
  "Returns a sequence which is a printable version of the node set NS."
  ns)
;
;
; =============================================================================
;
; is.nas
; ------
;
;       arguments     : u - <universal>
;
;       returns       : <boolean>
;
;       description   : It returns "true" if "u" is a <node-access set>,
;                       "false" otherwise.
;
;                                        written :  ejm 06/19/84
;                                        modified:
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun is.nas (u)
  "Returns T if U is a node-access set."
  (and (listp u) 
       (is1.nas u)))
;
;
; =============================================================================
;
; is1.nas
; -------
;
;       arguments     : s - <sequence>
;
;       returns       : <boolean>
;
;       description   : Auxiliary function to is.nas. It tests wether all the
;                       members of the <sequence> "u" are <node-access>es.
;
;                                        written :  ejm 06/19/84
;                                        translated: rmq 02/18/86
;                                        modified: scs 06/06/87
;
;
(defun is1.nas (s)
  "Returns T if every member of the sequence S is a node-access."
   (cond ((null s) t)
         ((is.n (node (first s))) (is1.nas (rest s)))
         ((and (numberp (first s)) (node (un-ize (first s))))
          (is1.nas (rest s)))
         (t nil)))
;
;
; =============================================================================
;
; apply-subst.ns
; --------------
;
;       arguments     : subst - <substitution>
;                       nodeset - <node set>
;
;       returns       : <node set>
;
;       description   : applies the <substitution> "subst" to the <node set>
;                       "nodeset" and returns the result
;
;                                        written :  rgh  8/21/85
;                                        modified:  rgh  4/11/86
;                                        modified:  scs  6/16/88
;
;
(defun apply-subst.ns (subst nodeset)
  (let ((resultns (new.ns)))
    (do.ns (n nodeset resultns)
       (setq resultns (insert.ns (match::applysubst n subst) resultns)))))
;
;
; =============================================================================
;
; cardinality.ns
; --------------
;
;       arguments     : ns - <node set>
;
;       returns       : <non-negative integer>
;
;       description   : returns the number of <node>s in "ns"
;
;                                        written :  rgh  4/06/86
;                                        modified:
;
;
(defmacro cardinality.ns (ns)
  `(length ,ns))
;
;
; =============================================================================

(defmacro remove-if-not.ns (test ns)
  "Returns a <node set> just like ns, but with all <nodes> failing the test removed."
  `(remove-if-not ,test ,ns))
